# -*-coding:utf-8-*-

__all__ = ['DateUtils']

import arrow


class DateUtils:
    arrow = arrow

    @staticmethod
    def arrow_str(arrow1):
        return arrow.get(arrow1).datetime.strftime("%Y-%m-%d")

    @staticmethod
    def days_in_week(arrow1):
        arrow1 = arrow.get(arrow1)
        return arrow.Arrow.range(
            'day',
            arrow1.floor('week'),
            arrow1.ceil('week'),)

    @staticmethod
    def weeks_in_range(arrow1, arrow2):
        return arrow.Arrow.range(
            'week',
            arrow.get(arrow1).floor('week'),
            arrow.get(arrow2).ceil('week'),)

    @staticmethod
    def fixed_weeks_in_range(date_range_str):
        """ 修复 一个范围内所有全部覆盖的weeks，即最坏情况是掐头去尾。"""
        # 兼容如果date_range的最后一个不是星期天，那该周日志就不完整。
        assert len(date_range_str) == 21  # e.g. "2014-09-01-2014-11-19"
        first_date = arrow.get(date_range_str[0:10])
        last_date = arrow.get(date_range_str[11:21])
        dates = DateUtils.weeks_in_range(first_date, last_date)
        if len(dates) > 0:
            if last_date.weekday() != 6:    # 6 index is Sunday
                dates = dates[:-1]
            if first_date.weekday() != 0:  # 0 index is Monday
                dates = dates[1:]
        return dates

    @staticmethod
    def date_value_by_type_in_last(date_value_1, date_type_1):
        val1 = arrow.get(date_value_1).replace(**{(date_type_1 + 's'): -1}) \
                                      .floor(date_type_1)
        return val1



# -*-coding:utf-8-*-

__all__ = ["TaskWeekHadoop"]

from .task_week import TaskWeek

import luigi


class TaskWeekHadoop(luigi.hadoop.HadoopExt, TaskWeek):

    def requires_with_prev_week(self, ref_task1):
        """ 依赖的天数据，以及上周的自身类统计数据 """
        total_tasks = [ref_task1(date1) for date1 in self.days_in_week]
        prev_week_stat_task1 = self.pre_task_by_self
        if isinstance(prev_week_stat_task1, self.task_class):
            total_tasks.append(prev_week_stat_task1)  # 如果不是 RootTask 的话
        return total_tasks



# -*-coding:utf-8-*-

__all__ = ["ref_tasks"]

from ..manager import load_a_task_by_name, luiti_config


def ref_tasks(*tasks):  # 装饰器
    """
    自动把依赖 Task 链接起来，通过属性访问。

    Example:

    ```python
    @ref_tasks("TaskB", "TaskC")
    class TaskA(TaskWeekBase):
        pass

    TaskA().TaskB == TaskB
    TaskA().TaskC == TaskC
    ```
    """
    def wrap_cls(ref_task_name):
        def _func(self):
            v1 = self.__dict__.get(ref_task_name, None)
            if v1 is None:
                v1 = load_a_task_by_name(ref_task_name)
                self.__dict__[ref_task_name] = v1
            return v1
        return _func

    def wrap_instance(ref_task_name, task_name):
        def _func(self):
            v1 = self.__dict__.get(task_name, None)
            if v1 is None:
                v1 = getattr(self, ref_task_name)(self.date_value)
                self.__dict__[task_name] = v1
            return v1
        return _func

    def __getstate__(self):
        """ Fix luiti_tasks module namespace conflicts. """
        for ref_task1 in self._ref_tasks:
            cname = ref_task1           # class    name
            iname = ref_task1 + "_task"  # instance name

            if cname in self.__dict__:
                del self.__dict__[cname]
            if iname in self.__dict__:
                del self.__dict__[iname]
        return self.__dict__

    def __setstate__(self, d1):
        # 1. default
        self.__dict__.update(d1)
        # 2. plug other package in `.__init_luiti`
        luiti_config.curr_project_name = self.package_name
        luiti_config.link_packages()

# cached_property 捕获不了 ref_task_name 变量, 被重置为某一个了。。
# property 可以捕获 ref_task_name 变量。
    def func(cls):
        setattr(cls, "_ref_tasks", tasks)
        for ref_task_name in cls._ref_tasks:
            setattr(cls, ref_task_name, property(wrap_cls(ref_task_name)))

            # TODO 根据当前日期返回。
            task_name = "%s_%s" % (ref_task_name, "task")
            setattr(cls, task_name,
                    property(wrap_instance(ref_task_name, task_name)))

            # clear ref task info when pickle.dump
            setattr(cls, "__getstate__", __getstate__)
        return cls
    return func



# -*-coding:utf-8-*-

__all__ = ['HadoopExt']

import luigi.hadoop
from .utils import ExtUtils, TargetUtils
from .parameter import ArrowParameter


class LuitiHadoopJobRunner(luigi.hadoop.HadoopJobRunner):

    """ overwrite DefaultHadoopJobRunner.class """

    # params are copied from HadoopJobRunner
    def __init__(self, libjars=None, output_format=None):
        config = luigi.hadoop.configuration.get_config()
        opts = {
            "streaming_jar": config.get('hadoop', 'streaming-jar'),
            "output_format": output_format,
            "libjars": libjars,
        }
        super(LuitiHadoopJobRunner, self).__init__(**opts)


class HadoopExt(luigi.hadoop.JobTask, ExtUtils.ExtendClass):

    run_mode = "mr_distribute"
    n_reduce_tasks = 1  # 体现在 输出的part-00000数量为reduce数量

    output_format = [
        # 单路输出。这个版本有问题。
        # "org.apache.hadoop.mapreduce.lib.output.TextOutputFormat",
        "org.apache.hadoop.mapred.TextOutputFormat",  # 单路输出
        "org.apache.hadoop.mapred.lib.MultipleTextOutputFormat",  # 多路输出
    ][0]  # 默认是 单路输出
    output_format_default = output_format[:]
    libjars = []

    def __init__(self, *args, **kwargs):
        """ 参考 TaskBase, 确保在 继承时还可以有TaskBase的覆写日期功能。 """
        super(HadoopExt, self).__init__(*args, **kwargs)
        self.orig_date_value = self.orig_date_value or \
            ArrowParameter.get(self.date_value)
        self.reset_date()

    # overwrite
    def job_runner(self):
        """ will be wraped in `run` function. """
        # Auto compile java code
        if self.output_format != self.output_format_default:
            self.compile_java_code()

        return LuitiHadoopJobRunner(
            output_format=self.output_format, libjars=self.libjars)

    def output(self):
        return TargetUtils.hdfs(self.data_file)

    def jobconfs_opts(self):
        return [
            "mapreduce.framework.name=yarn",
            'mapred.reduce.tasks=%s' % self.n_reduce_tasks,
        ]

    def jobconfs(self):
        jcs = super(luigi.hadoop.JobTask, self).jobconfs()
        for conf_opt_1 in self.jobconfs_opts():
            jcs.append(conf_opt_1)
        return jcs

    # TestCase related attrs
    def mrtest_input(self):
        raise NotImplementedError

    def mrtest_output(self):
        raise NotImplementedError

    def mrtest_attrs(self):
        return dict()



# -*-coding:utf-8-*-

__all__ = ["persist_files"]

import os
from luigi import Event
from ..utils import IOUtils


def persist_files(*files):  # 装饰器
    """ 多个data_file 可以用 DSL 描述，然后和 event_handler(Event.FAILURE) 绑定在一起 """
    def func(cls):
        # 1. 设置 持久化文件属性
        def wrap(file1):  # 这样才可以保存 file1 变量，而不至于被覆写。
            def _file(self):
                return os.path.join(self.data_dir, file1 + ".json")
            return _file

        setattr(cls, "__persist_files", files)
        for file1 in getattr(cls, "__persist_files"):
            setattr(cls, file1, property(wrap(file1)))  # @decorator

        # 2. 绑定 失败时删除这些文件
        def clean_tmp(task, exception):
            for file1 in files:
                IOUtils.remove_files(getattr(task, file1))
            # IOUtils.remove_files(task.data_file)
            # NOTE 好像 Hadoop 会自动处理失败任务的输出文件的，否则就会导致其在N次重试一直在running。
        cls.event_handler(Event.FAILURE)(clean_tmp)

        return cls

    return func



# -*-coding:utf-8-*-

from ..time.task_base import TaskBase, cached_property
from etl_utils import process_notifier
from ...utils import CommandUtils, TargetUtils, MRUtils, HDFSUtils
import luigi
import os
import arrow
import json


class MongoImportTask(TaskBase):

    report_status_collection_name = "report_status"
    report_status_namespace = "latestCollection"
    report_name = NotImplementedError

    system_tmp = "/tmp"  # default

    @cached_property
    def mongodb_db(self):
        return self.mongodb_connection[self.database_name]

    @cached_property
    def report_status_collection_model(self):
        return self.mongodb_db[self.report_status_collection_name]

    @cached_property
    def data_file_collection_model(self):
        return self.mongodb_db[self.collection_name]

    # 1. config
    @cached_property
    def source_task(self):
        raise NotImplementedError

    @cached_property
    def mongodb_connection(self):
        raise NotImplementedError

    @cached_property
    def database_name(self):
        raise NotImplementedError

    @cached_property
    def index_schema(self):
        raise NotImplementedError

    def convert_line_to_json(self, line1):
        classid_uid_1, val_1 = MRUtils.split_mr_kv(line1)
        return val_1

    def run_before_hook(self):
        pass

    def run_after_hook(self):
        pass

    def find_invalid_record(self, stat_1):
        return False

    # 2. common
    def requires(self):
        return [getattr(self, _ref_task_1)(self.date_value)
                for _ref_task_1 in self._ref_tasks]

    def run(self):
        self.run_before_hook()

        # 1. check is already done.
        if self.is_collection_exists():
            return False

        # 2. check report status collection is valid
        if self.report_status_collection_model.count() == 0:
            self.report_status_collection_model.insert(
                {self.report_status_namespace: {}})
        assert self.report_status_collection_model.count() == 1, "更新纪录 只能有一条！"

        # 3. output json with err
        data_file1 = self.source_task_instance.data_file
        source1 = luigi.HDFS(data_file1)
        tmp_file1 = open(self.tmp_filepath, 'w')
        tmp_errfile1 = open(self.tmp_errfilepath, 'w')

        for line1 in process_notifier(
                TargetUtils.line_read(source1), u"[read lines] %s" % source1):
            if len(line1) == 0:
                continue  # meaningless data.
            stat_2 = self.convert_line_to_json(line1)
            if self.find_invalid_record(stat_2):
                tmp_errfile1.write(line1 + "\n")
            else:
                tmp_file1.write(json.dumps(stat_2) + "\n")
        tmp_file1.close()
        tmp_errfile1.close()

        # 4. upload to mongodb
        HDFSUtils.copyFromLocal(self.tmp_errfilepath, self.data_file_err)
        CommandUtils.execute(self.mongo_ensure_index)
        CommandUtils.execute(self.mongoimport_command)

        # 5. clean tmp
        CommandUtils.execute("rm -f %s" % self.tmp_filepath)
        CommandUtils.execute("rm -f %s" % self.tmp_errfilepath)

        # 6. update report status
        item1 = self.report_status_collection_model.find()[0]
        del item1['_id']
        item1[self.report_status_namespace][self.report_name] = {
            'collection_name': self.collection_name,
            'updated_at': arrow.now().datetime,
        }
        self.report_status_collection_model.find_and_modify(
            query={},
            update={"$set": item1},
            full_response=True
        )

        self.run_after_hook()

        return True

    def is_collection_exists(self):
        return self.data_file_collection_model.count() > 0

    @cached_property
    def source_task_instance(self):
        return self.source_task(self.date_value)

    @cached_property
    def mongoimport_command(self):
        return "/usr/bin/mongoimport " + \
            ("--host %s " % self.mongodb_connection.host) + \
            ("--port %s " % self.mongodb_connection.port) + \
            ("--db %s " % self.database_name) + \
            ("--collection %s " % self.collection_name) + \
            ("--file %s " % self.tmp_filepath)

    @cached_property
    def mongo_ensure_index(self):
        if not isinstance(self.index_schema, (str, unicode)):
            self.index_schema = json.dumps(self.index_schema)
        js_str = "db.%s.ensureIndex(%s)" % \
            (self.collection_name, self.index_schema)
        return self.mongo_eval(js_str)

    def mongo_eval(self, js_str):
        return "/usr/bin/mongo " + \
            ("%s:%s/%s " % (self.mongodb_connection.host, self.mongodb_connection.port, self.database_name)) + \
            ("--eval \"%s\" " % js_str)

    @cached_property
    def collection_name(self):
        """ e.g. redmine5954_parent_report_week_20140901 """
        return self.data_name + "_" + self.date_value.strftime("%Y%m%d")

    @cached_property
    def tmp_filepath(self):
        return self.tmp_dir + "/" + self.date_value.strftime("%Y%m%d")

    @cached_property
    def tmp_errfilepath(self):
        return self.tmp_filepath + ".err"

    @cached_property
    def data_file_err(self):
        return self.data_file + ".err"

    @cached_property
    def tmp_dir(self):
        dir1 = self.system_tmp + "/" + self.task_class.__name__
        os.system("mkdir -p %s" % dir1)
        return dir1



# -*-coding:utf-8-*-

__all__ = ["luigi"]


import luigi
import luigi.hadoop
luigi.hadoop = luigi.hadoop
import luigi.hdfs
luigi.hdfs = luigi.hdfs

from ..hadoop_ext import HadoopExt
luigi.hadoop.HadoopExt = HadoopExt  # write back
# NOTE 对 luigi.hadoop 兼容 "track the job: "


import os
from ..utils import TargetUtils
luigi.HDFS = TargetUtils.hdfs  # 本来就是需要读取全局配置，所以索性就绑定在 luigi 命名空间了吧。


from .check_date_range import check_date_range
from .check_runtime_range import check_runtime_range
from .mr_local import mr_local
from .multiple_text_files import multiple_text_files
from .persist_files import persist_files
from .plug_packages import plug_packages
from .ref_tasks import ref_tasks

luigi.persist_files = persist_files
luigi.ref_tasks = ref_tasks
luigi.check_date_range = check_date_range
luigi.check_runtime_range = check_runtime_range
luigi.mr_local = mr_local
luigi.multiple_text_files = multiple_text_files
luigi.plug_packages = plug_packages


orig_create_packages_archive = luigi.hadoop.create_packages_archive


def create_packages_archive_with_support_egg(packages, filename):
    """
    Fix original luigi's `create_packages_archive` cannt attach egg packages
    (zip file type) to tarfile, Cause it's coping file mechanism by absolute
    path.
    """
    # 1. original create tar file
    orig_create_packages_archive(packages, filename)

    # 2. append python egg packages that 1. not covered
    import tarfile
    tar = tarfile.open(filename, "a")

    logger = luigi.hadoop.logger
    fake_exists_path = "/"  # root is awlays exists

    def get_parent_zip_file_within_absolute_path(path1):
        path2 = path1[:]
        is_success = False
        while path2 != fake_exists_path:
            path2 = os.path.dirname(path2)
            if os.path.isfile(path2):
                is_success = True
                break
        return is_success, path2

    def add(src, dst):
        logger.debug('adding to tar: %s -> %s', src, dst)
        tar.add(src, dst)

    import zipfile
    import tempfile
    for package1 in packages:
        path2 = (getattr(package1, "__path__", []) + [fake_exists_path])[0]
        if os.path.exists(path2):
            continue  # so luigi can import it.
        if not path2.startswith("/"):
            continue  # we only care about libraries.

        is_success, zipfilename3 = \
            get_parent_zip_file_within_absolute_path(path2)
        if is_success:
            tmp_dir3 = tempfile.mkdtemp()
            zipfile.ZipFile(zipfilename3).extractall(tmp_dir3)

            for root4, dirs4, files4 in os.walk(tmp_dir3):
                for file5 in files4:
                    if file5.endswith(".pyc"):
                        continue
                    add(
                        os.path.join(root4, file5),
                        os.path.join(
                            root4.replace(tmp_dir3, "").lstrip("/"), file5))

    client_cfg = os.path.join(os.getcwd(), "client.cfg")
    if os.path.exists(client_cfg):
        tar.add(client_cfg, "client.cfg")
    tar.close()
luigi.hadoop.create_packages_archive = create_packages_archive_with_support_egg


from ..manager import luiti_config, active_packages
luigi.ensure_active_packages = lambda: active_packages  # make a wrap
luigi.luiti_config = luiti_config
luiti_config.linked_luigi = luigi



# -*-coding:utf-8-*-

__all__ = ["check_date_range"]

from ..parameter import ArrowParameter


def check_date_range():  # 装饰器
    """
    从数据库导数据时，必须注意时间范围内的所有数据是否都齐全了。如果未齐全，
    即在当前时间范围里导的话，那么就会缺失数据了，相当于提前导了。

    比如在周六就把这周的关联数据导出来，那么周日的数据就没包含在里面。应该在下周一后才开始导。
    """
    def decorator(orig_run):
        def new_run(self):
            # 说明时间未到，然后就直接退出
            if self.date_value_by_type_in_begin <= \
                    ArrowParameter.now() < self.date_value_by_type_in_end:
                return False
            return orig_run(self)
        return new_run

    def func(cls):
        cls.run = decorator(cls.run)
        return cls
    return func
# TODO support Hadoop



# -*-coding:utf-8-*-

import os
import glob
from inflector import Inflector
from etl_utils import singleton, cached_property

from .config import luiti_config as lc
from .active_packages import active_packages


@singleton()
class PackageMapClass(object):

    @cached_property
    @active_packages
    def task_clsname_to_package(self):

        assert lc.luiti_tasks_packages, "At least have one project!"

        result = dict()
        for project1 in lc.luiti_tasks_packages:
            project_dir2 = project1.__path__[0]

            # if it's not a zip file, but a normal package directory
            is_zip_file = os.path.exists(
                os.path.join(project_dir2, "__init__.py"))
            if not is_zip_file:
                raise Exception(
                    """[setup.py format error] make sure """
                    """project "%s" zip_safe=False option exists!"""
                    % project1.__name__)

            task_path_pattern = os.path.join(
                project_dir2, "luiti_tasks/[a-z]*.py")

            for f2 in glob.glob(task_path_pattern):
                task_filename3 = os.path.basename(f2).rsplit(".", 1)[0]
                task_clsname4 = Inflector().classify(task_filename3)
                result[task_clsname4] = project1
        return result
PackageMap = PackageMapClass()



# -*-coding:utf-8-*-

__all__ = [
    "IOUtils",
    "TargetUtils",
    "MRUtils",
    "MathUtils",
    "HDFSUtils",
    "CommandUtils",
    "CompressUtils",
    "DateUtils",
    "ExtUtils"
]

from .io_utils import IOUtils
from .target_utils import TargetUtils
from .mr_utils import MRUtils
from .math_utils import MathUtils
from .hdfs_utils import HDFSUtils
from .command_utils import CommandUtils
from .date_utils import DateUtils
from .compress_utils import CompressUtils
from .ext_utils import ExtUtils



# -*-coding:utf-8-*-

import json
import luigi
import luigi.hdfs
from luigi import LocalTarget
from etl_utils import JsonUtils
from .target_utils import TargetUtils


class IOUtils:

    SQL_RANGE_LIMIT = 1000

    @staticmethod
    def write_json_to_output(json1, output1):
        with output1.open('w') as output_hdfs:
            m1 = lambda item1: json.dumps(list(item1))
            m2 = lambda item1: JsonUtils.unicode_dump(item1).encode("UTF-8")
            if isinstance(json1, (list, set,)):
                # 兼容 JsonUtils.unicode_dump 不支持list
                method = m1
            else:
                method = m2
            output_hdfs.write(method(json1) + "\n")

    @staticmethod
    def read_json_from_output(output1):
        # only one line
        item1 = None
        read_line_count = 0
        for json1 in TargetUtils.json_read(output1):
            read_line_count += 1
            item1 = json1
            if read_line_count >= 2:
                raise ValueError("[multiple line error]"
                                 " %s should contain only one line!" % output1)
        return item1

    @staticmethod
    def remove_files(*files):  # 兼容 写入中途失败
        for file1 in files:
            if luigi.hdfs.exists(file1):
                luigi.hdfs.remove(file1)

    @staticmethod
    def local_target(path1):
        return LocalTarget(path1)



# -*-coding:utf-8-*-

__all__ = ['luigi', 'config',

           'TaskBase',
           'TaskHour',
           'TaskDay',
           'TaskDayHadoop',
           'TaskWeek',
           'TaskWeekHadoop',
           'TaskMonth',
           'TaskRange',
           'TaskRangeHadoop',

           'RootTask',
           'StaticFile',
           'MongoImportTask',
           'HadoopExt',

           'manager',

           'IOUtils', 'DateUtils', 'TargetUtils', 'HDFSUtils',
           'MRUtils', 'MathUtils', 'CommandUtils',
           'CompressUtils',

           'ArrowParameter',

           'os', 're', 'sys', 'defaultdict', 'json', 'cached_property',
           'arrow',

           'MrTestCase', ]

import os
import sys
import re
from collections import defaultdict
import json
from etl_utils import cached_property

from .luigi_decorators import luigi
from .hadoop_ext import HadoopExt
from .task_templates.time.task_base import TaskBase
from .task_templates.time.task_hour import TaskHour
from .task_templates.time.task_day import TaskDay
from .task_templates.time.task_week import TaskWeek
from .task_templates.time.task_month import TaskMonth
from .task_templates.time.task_range import TaskRange

from .task_templates.time.task_day_hadoop import TaskDayHadoop
from .task_templates.time.task_week_hadoop import TaskWeekHadoop
from .task_templates.time.task_range_hadoop import TaskRangeHadoop

from .task_templates.other.root_task import RootTask
from .task_templates.other.static_file import StaticFile
from .task_templates.other.mongo_import_task import MongoImportTask

from . import manager
from .utils import IOUtils, DateUtils, TargetUtils, HDFSUtils, \
    MRUtils, MathUtils, CommandUtils, CompressUtils

from .parameter import ArrowParameter, arrow

from .mr_test_case import MrTestCase


# TODO 整理

config = manager.luiti_config

luigi.tmp_dir = "/tmp"  # default one



# -*-coding:utf-8-*-

__all__ = [
    "PackageMap",
    "SysArgv",
    "ld",
    "load_a_task_by_name",
    "print_all_tasks",
    "new_a_project",
    "generate_a_task",
    "find_dep_on_tasks",
    "Table",
    "get_all_date_file_to_task_instances",
    "print_files_by_task_cls_and_date_range",
    "soft_delete_files",
    "active_packages",
]

from .loader import Loader
from .table import Table
from .dep import Dep
from .files import Files

from .config import luiti_config
from .package_map import PackageMap
from .active_packages import active_packages

from .sys_argv import SysArgv

from .generate_from_templates import GenerateFromTemplates


# API list

find_dep_on_tasks = Dep.find_dep_on_tasks
get_all_date_file_to_task_instances = Files.get_all_date_file_to_task_instances
soft_delete_files = Files.soft_delete_files
load_all_tasks = Loader.load_all_tasks
load_a_task_by_name = Loader.load_a_task_by_name
import2 = luiti_config.import2
print_all_tasks = Table.print_all_tasks
print_files_by_task_cls_and_date_range = \
    Table.print_files_by_task_cls_and_date_range
new_a_project = GenerateFromTemplates.new_a_project
generate_a_task = GenerateFromTemplates.generate_a_task


from etl_utils import singleton, cached_property


@singleton()
class LazyData(object):

    @cached_property
    def all_task_classes(self):
        return [i1['task_cls'] for i1 in ld.result['success']]

    @cached_property
    def result(self):
        return load_all_tasks()

ld = LazyData()



# -*-coding:utf-8-*-

from ..parameter import ArrowParameter
import luigi.hdfs
from datetime import datetime


class Files(object):

    """ Get all outputs which generated by luiti tasks. """

    @staticmethod
    def get_all_date_file_to_task_instances(date_range, task_classes):
        """ return all instances in date range. """
        assert_msg = "[error] correct format is \"20140901-20140905\", " \
                     "but the input is %s" % date_range
        assert len(date_range) == 17, assert_msg

        first_date, last_date = date_range[0:8], date_range[9:]
        first_date, last_date = ArrowParameter.get(
            first_date, "YYYYMMDD"), ArrowParameter.get(last_date, "YYYYMMDD")

        return dict({file_3: task_instance_2
                     for task1 in task_classes
                     for task_instance_2 in task1.instances_by_date_range(
                         first_date, last_date)
                     for file_3 in task_instance_2._persist_files +
                     [task_instance_2.data_file]})

    @staticmethod
    def soft_delete_files(*files):
        delete_at_str = datetime.now().strftime("-deleted-at-%Y%m%d-%H%M%S")

        for file1 in sorted(files):
            print "[delete file]", file1
            if luigi.hdfs.client.exists(file1):
                luigi.hdfs.client.rename(file1, file1 + delete_at_str)
                print
            else:
                print "[err] doesnt exist!"

        print "\nDone!"
        return 0



# -*-coding:utf-8-*-

__all__ = ["mr_local"]

from collections import defaultdict
from etl_utils import process_notifier
from ..utils import TargetUtils


def mr_local(**opts):
    """
    Sometimes Hadoop streaming sucks, so we only use the solid HDFS, and turn
    MapReduce job into local mode.

    And `mr_local` is optimized by a fixed chunk write operation.
    """

    def mr_run(self):
        """ Overwrite BaseHadoopJobTask#run function. """
# TODO maybe model cache
        map_kv_dict = defaultdict(list)

        inputs = self.input()
        if not isinstance(inputs, list):
            inputs = [inputs]
        for input_hdfs_1 in inputs:
            for line2 in TargetUtils.line_read(input_hdfs_1):
                for map_key_3, map_val_3 in self.mapper(line2):
                    map_kv_dict[map_key_3].append(map_val_3)

        with self.output().open("w") as output1:
            fixed_chunk = list()
            for reduce_key_2 in process_notifier(map_kv_dict.keys()):
                reduce_vals_2 = map_kv_dict[reduce_key_2]
                for _, reduce_val_2 in self.reducer(
                        reduce_key_2, reduce_vals_2):
                    fixed_chunk.append(reduce_val_2)

                    if len(fixed_chunk) % self.chunk_size == 0:
                        output1.write("\n".join(fixed_chunk) + "\n")
                        fixed_chunk = list()
                del map_kv_dict[reduce_key_2]
            output1.write("\n".join(fixed_chunk) + "\n")

    def wrap(cls):
        cls.run = mr_run
        cls.run_mode = "mr_local"

        opts["chunk_size"] = opts.get("chunk_size", 100)
        for k1, v1 in opts.iteritems():
            setattr(cls, k1, v1)

        return cls
    return wrap



# -*-coding:utf-8-*-


from setuptools import setup

setup(
    name='luiti',
    version='0.1.3',
    url='http://github.com/17zuoye/luiti/',
    license='MIT',
    author='David Chen',
    author_email=''.join(reversed("moc.liamg@emojvm")),
    description='Luiti = Luigi + time',
    long_description=open("README.markdown").read(),
    packages=[
                'luiti',
                'luiti/luigi_decorators',
                'luiti/manager',
                'luiti/task_templates/',
                'luiti/task_templates/time',
                'luiti/task_templates/other',
                'luiti/utils', ],
    scripts=[
        'bin/luiti',
    ],

    package_data={'luiti': ["luiti/java/*.java"]},
    include_package_data=True,

    zip_safe=False,
    platforms='any',
    install_requires=[
        "luigi         == 1.1.2",
        "snakebite     == 2.5.2",

        "arrow         == 0.4.4",
        "inflector     == 2.0.11",
        "protobuf      == 2.6.1",
        "tornado       == 4.0.2",
        "mechanize     == 0.2.5",

        "python-daemon == 1.6.1",
        "MySQL-python  >= 1.2.5",
        "pymongo       >= 2.7.2",

        "etl_utils     == 0.1.10",
        "tabulate      == 0.7.3",
        "pipetools     == 0.2.7",
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*-coding:utf-8-*-

__all__ = ['TaskWeek']

from .task_base import TaskBase, cached_property
from ...utils import DateUtils


class TaskWeek(TaskBase):

    @cached_property
    def days_in_week(self):
        return list(DateUtils.days_in_week(self.date_value))



# -*- coding: utf-8 -*-

__all__ = ['MrTestCase']


from collections import defaultdict
import json

from .manager import Loader


def MrTestCase(cls, verbose=False):
    """
    功能: 集成测试数据到 类中 ，这样就方便引用了。
    """

    assert "mr_task_names" in dir(cls), "%s must assgin some task names!" % cls

    cls.maxDiff = None  # compact large json diff

    def map_lines(text):
        assert isinstance(text, unicode)
        result = list()
        for l1 in text.split("\n"):
            l1 = l1.strip()
            if not l1:
                continue
            result.append(l1)
        return result

    def generate_closure_function(mr_task_name1):
        task_cls = Loader.load_a_task_by_name(mr_task_name1)  # keep it!
        if verbose:
            print "[task_cls]", task_cls

        def test_mr(self):
            task_instance_1 = task_cls("2014-09-01")
            if verbose:
                print "[task_instance]", task_instance_1

            task_instance_1.lines = map_lines(task_instance_1.mrtest_input())
            result_expect = sorted(
                [json.loads(i2) for i2
                 in map_lines(task_instance_1.mrtest_output())])

            self.assertEqual(result_expect, run_map_reduce(task_instance_1))
        return test_mr

    for mr_task_name1 in cls.mr_task_names:
        test_method_name = "test_" + mr_task_name1
        if verbose:
            print
        if verbose:
            print "[test_method_name]", test_method_name

        setattr(
            cls, test_method_name, generate_closure_function(mr_task_name1))

        if verbose:
            print
        if verbose:
            print

    return cls


def run_map_reduce(task_instance_1):
    # 1. bind attrs
    for k1, v1 in task_instance_1.mrtest_attrs().iteritems():
        setattr(task_instance_1, k1, v1)

    # 2. map it!
    mapper_key_to_vals = defaultdict(list)
    for line1 in task_instance_1.lines:
        for key_1, val_1 in task_instance_1.mapper(line1.strip()):
            mapper_key_to_vals[key_1].append(val_1)

    # 3. reduce it!
    result_list = list()
    for key_1, vals_1 in mapper_key_to_vals.iteritems():
        vals_generator = iter(vals_1)
        for _, val_2 in task_instance_1.reducer(key_1, vals_generator):
            result_list.append(json.loads(val_2))
    return sorted(result_list)



# -*-coding:utf-8-*-

__all__ = ['TaskRange']

from .task_base import TaskBase
from ...utils import DateUtils


class TaskRange(TaskBase):

    # NOTE date_value 和 date_range 两个值是必须的。
    # 1. date_value 是写到那个日期目录
    # 2. date_range 是指定了依赖的日期范围

    def date_range(self):
        raise ValueError("Overwrite Me!")
    # date_range = luigi.DateIntervalParameter()
    # date_range = luigi.Parameter() # 临时现为 str 类型吧

    @property
    def dates_in_range(self):
        # method_1 = self.date_type + "s_in_range" # e.g. weeks_in_range
        method_1 = 'week' + "s_in_range"  # NOTE 目前直接为 week, 因为是range.

# s1 = "2014-10-01-2014-10-07"
# s1[0:10]  => '2014-10-01'
# s1[11:21] => '2014-10-07'
        date_1, date_2 = self.date_range[0:10], self.date_range[11:21]

        return list(getattr(DateUtils, method_1)(date_1, date_2))



# -*-coding:utf-8-*-

import sys


class SysArgv(object):

    @staticmethod
    def convert_to_luigi_accepted_argv(subparsers):
        sys.argv_old = sys.argv[:]
        luigi_keep_opts = ["--date-value"]

        def fetch_keys(parser1):
            return parser1.__dict__['_option_string_actions'].keys()

        luiti_only_opts = subparsers.choices.keys() + \
            list(set(
                [k3 for p2 in subparsers._name_parser_map.values()
                    for k3 in fetch_keys(p2)]))
        luiti_only_opts = [i1 for i1 in luiti_only_opts
                           if i1 not in luigi_keep_opts]

        delete_argv_idxes = set([])
        for idx1, arg1 in enumerate(sys.argv):
            if idx1 in delete_argv_idxes:
                continue
            # 1. 排除 tasks, files, run 等
            if (not arg1.startswith("--")) and (arg1 in luiti_only_opts):
                delete_argv_idxes.add(idx1)
                continue
            # 2. 处理 --task-name 等参数
            if "=" in arg1:
                arg2, val2 = arg1.split("=", 1)
                if arg2 in luiti_only_opts:
                    delete_argv_idxes.add(idx1)
            else:
                if (arg1 in luiti_only_opts) and (arg1 not in luigi_keep_opts):
                    delete_argv_idxes.add(idx1)
                    delete_argv_idxes.add(idx1 + 1)
        sys.argv = [arg1 for idx1, arg1 in enumerate(sys.argv)
                    if idx1 not in delete_argv_idxes]



# -*-coding:utf-8-*-

__all__ = ["multiple_text_files"]

import os
import commands
from ..utils import CommandUtils


def multiple_text_files(task_cls):
    """
    让当前 MapReduce 任务的结果可以输出到多个文件。

     使用说明:

    ```python
    @luigi.MultipleTextOutputFormat
    class ManAndWomanDay(TaskDayHadoop):
        def mapper(self, line1):
            item1 = MRUtils.json_parse(line1)
            yield item1['uid'], item1

        def reducer(self, uid1, vals_1):
            for item1 in vals_1:
                yield item1["gender"], MRUtils.str_dump(item1)
    ```

    以上代码就把男性和女性分流到两个文件了。文件名类似
    1. man_and_woman_day.json/man
    2. man_and_woman_day.json/woman

    而不是默认输出的
    1. man_and_woman_day.json/part-00000

    WARN:
        when use `@luigi.multiple_text_files`, consider to wrap subfolders with
        StaticFile task class.
    """

    java_namespace = "com.voxlearning.bigdata.MrOutput"
    java_lib = "MultipleTextFiles"
    output_format = ".".join([java_namespace, java_lib])

    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    java_file = java_lib + ".java"
    target_class = java_lib + ".class"
    target_jar = os.path.join(root_dir, "java", java_lib + ".jar")

    def compile_java_code(self):
        """ compile java code dynamically. """
        if os.path.exists(target_jar):
            return False

        classes_dir = java_namespace.replace(".", "/")
        javac_cmd = commands.getoutput("which javac")
        java_classpath = commands.getoutput("hadoop classpath")
        jar_cmd = commands.getoutput("which jar")

        compile_cmd = ";\n".join([
            # no absolute path, compact with java namespace.
            "cd %s/java" % root_dir,

            """%s -classpath "%s" %s""" % (javac_cmd,
                                           java_classpath, java_file, ),
            "rm -rf %s" % classes_dir,
            "mkdir -p %s" % classes_dir,
            "cp %s %s" % (target_class, classes_dir),
            "%s cvf %s %s/*.class" % (jar_cmd, target_jar, classes_dir, ),
        ])

        CommandUtils.execute(compile_cmd)

    setattr(task_cls, "output_format", output_format)
    setattr(task_cls, "libjars", [target_jar, ])
    setattr(task_cls, "compile_java_code", compile_java_code)
    return task_cls



# -*-coding:utf-8-*-


from ...luigi_decorators import luigi
from ...utils import TargetUtils


class StaticFile(luigi.Task):
    filepath = None

    # Mimic default luigi.ExternalTask
    def run(self):
        pass

    def complete(self):
        return True

    def output(self):
        assert self.filepath, u"Please assign `filepath` !"
        return TargetUtils.hdfs(self.filepath)



# -*-coding:utf-8-*-

import os


class Table(object):

    @staticmethod
    def puts(task_table, task_headers, **opts):
        from tabulate import tabulate
        print
        print tabulate(task_table, task_headers, **opts)
        print

    @staticmethod
    def print_all_tasks(result):
        """ input from Loader.load_all_tasks """
        def task_cls_inspect(cls):
            return cls.__name__

        task_headers = ["", "All Tasks"]
        task_table = [[idx1 + 1, task_cls_inspect(item1['task_cls'])]
                      for idx1, item1 in enumerate(sorted(result['success']))]
        task_table.extend([["total", len(result['success'])]])

        Table.puts(task_table, task_headers, tablefmt="grid")

        if result['failure']:
            print
            print "[warn] failure parsed files"
            print
            for failure1 in result['failure']:
                print "[task_file] ", failure1['task_clsname']
                print "[err] ", failure1['err']
                print

    @staticmethod
    def print_files_by_task_cls_and_date_range(curr_task, args, opts=None):
        opts = opts or dict()
        # 打印 依赖类 和 执行配置 信息
        task_headers = ["Current Env Key", "Current Env Value"]
        task_table = [
            ["task name", args.task_name],
            ["task date range", args.date_range],
            ["task execute mode", "DRY=" + str(args.dry)],
            ["task dep mode", "DEP=" + str(args.dep)],
            ["related task classes total count", opts['task_classes_count']],
        ]
        print
        print "Tasks related infos"
        Table.puts(task_table, task_headers, tablefmt="grid")

        # 打印 要删除的文件列表
        file_headers = ["Generated from task", "Storage",
                        "Date value", "Filename"]

        dep_file_to_task_instances = opts['dep_file_to_task_instances']
        file_table = [
            [dep_file_to_task_instances[f1].__class__.__name__,
             'HDFS', dep_file_to_task_instances[f1].date_str,
             os.path.basename(f1), ]
            for f1 in sorted(dep_file_to_task_instances.keys())]
        file_table.append(
            ['', '', '', "Total count %s" % len(dep_file_to_task_instances)])
        file_table.append(['', '', '', ''])
        file_uniq_root_dir = set(
            [t1.root_dir for t1 in opts['dep_tasks_on_curr_task']])
        file_table.append(
            ['All root dirs', '', '',
             'Total count %s' % len(file_uniq_root_dir)])
        for dir1 in file_uniq_root_dir:
            file_table.append(['', '', '', dir1])

        print
        print "Files related infos"
        Table.puts(file_table, file_headers, tablefmt="grid")
        print "\n" * 3

    @staticmethod
    def print_task_info(curr_task, dep_tasks_on_curr_task):
        task_headers = ["Task name", curr_task.__name__]
        task_content = [
            ["Tasks self dep on", str(list(curr_task._ref_tasks))],
            ["Tasks dep on self",
             str(sorted([t2.__name__ for t2 in dep_tasks_on_curr_task]))],
        ]
        Table.puts(task_content, task_headers, tablefmt="grid")
